import os
import timm
import torch
from torch.utils.tensorboard import SummaryWriter
from copy import deepcopy
from wildlife_tools.tools import realize


class EpochCheckpoint():
    def __init__(self, folder='.', save_step=1):
        self.folder = folder
        self.save_step = save_step

    def __call__(self, trainer, **kwargs):
        os.makedirs(self.folder, exist_ok=True)
        if trainer.epoch % self.save_step == 0:
            trainer.save(folder=self.folder, file_name=f"checkpoint-{trainer.epoch}.pth")


class EpochLog():
    def __init__(self, folder='.'):
        self.folder = folder
        self.writer = SummaryWriter(log_dir=folder)

    def __call__(self, trainer, epoch_data: dict[str, int], **kwargs):
        if trainer.scheduler is not None:
            self.writer.add_scalar('lr', trainer.scheduler.get_last_lr()[0], trainer.epoch)
        
        for key, value in epoch_data.items():
            self.writer.add_scalar(key, value, trainer.epoch)


class EpochCallbacks():
    ''' Returns sequence of epoch callbacks.'''

    def __init__(self, steps):
        self.steps = steps


    def __call__(self, trainer, **kwargs):
        for step in self.steps:
            step(trainer=trainer, **kwargs) 


    @classmethod
    def from_config(cls, config):
        config = deepcopy(config)
        steps = []
        for config_step in config['steps']:
            steps.append(realize(config_step))
        return cls(steps=steps)



import os
import pickle


class FeatureExtractor():
    def __call__(self, dataset):
        raise NotImplemented()


    def run_and_save(self, dataset, save_path):
        features = self(dataset)

        os.makedirs(save_path, exist_ok=True)
        name = self.__class__.__name__
        data = {
            'name': name,
            'features': features,
            'metadata': getattr(dataset, 'metadata', None),
        }

        file_name = os.path.join(save_path, name + '.pkl')
        with open(file_name, 'wb') as file:
            pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)
        return file_name



import cv2
import numpy as np
from tqdm import tqdm
import itertools
import faiss
from wildlife_tools.similarity.base import Similarity


def get_faiss_index(d, device='cpu'):
    if device == 'cuda':
        resource = faiss.StandardGpuResources()
        config = faiss.GpuIndexFlatConfig()
        config.device = 0
        return faiss.GpuIndexFlatL2(resource, d, config)
    elif device == 'cpu':
        return faiss.IndexFlatL2(d)
    else:
        raise ValueError(f'Invalid device: {device}')


class MatchDescriptors(Similarity):
    def __init__(
        self,
        descriptor_dim: int,
        thresholds: tuple[float] = (0.5, ),
        device: str = 'cpu',
    ):

        self.descriptor_dim = descriptor_dim
        self.thresholds = thresholds
        self.device = device


    def __call__(self, query, database):
        iterator = itertools.product(enumerate(query), enumerate(database))
        iterator_size = len(query)*len(database)
        similarities = {t: np.full((len(query), len(database)), np.nan, dtype=np.float16) for t in self.thresholds}

        index = get_faiss_index(d=self.descriptor_dim, device=self.device)
        for pair in tqdm(iterator, total=iterator_size, mininterval=1, ncols=100):
            (q_idx, q_data), (d_idx, d_data) = pair

            if (q_data is None) or (d_data is None):
                for t in self.thresholds:
                    similarities[t][q_idx, d_idx] = 0

            else:
                index.reset()
                index.add(q_data)
                score, idx = index.search(d_data, k=2)
                with np.errstate(divide='ignore'):
                    ratio = score[:, 0] / score[:, 1]
                for t in self.thresholds:
                    similarities[t][q_idx, d_idx] = np.sum(ratio < t)

        return similarities



import timm

class TimmBackbone():
    '''
    Example config in YAML file:
    
    backbone:
        method: TimmBackbone
        model_name: 'efficientnet_b0'
        pretrained: True
    '''

    @classmethod
    def from_config(cls, config, output_size=None):
        if output_size is None:
            output_size = 0
        return timm.create_model(num_classes=output_size, **config)



import timm
import torchvision.transforms as T


class TransformTimm():
    '''
    Pytorch transform function from timm library.
    Example configs in YAML file:

    transform:
        method: TransformTimm
        input_size: 224
        is_training: True
        auto_augment: 'rand-m10-n2-mstd1'
    '''

    @classmethod
    def from_config(cls, config):
        return timm.data.transforms_factory.create_transform(**config)


class TransformTorchvision():
    '''
    Pytorch transform function from torchvision library.
    Example configs in YAML file:

    transform:
        method: TransformTorchvision
        compose:
            - 'Resize(size=256)'
            - 'ToTensor()'
    '''

    @classmethod
    def from_config(cls, config):
        compose = config['compose']
        return T.Compose([eval(s, globals(), T.__dict__) for s in compose])



import os
import torch
import torch.nn.functional as F
from wildlife_tools.similarity.base import Similarity


def cosine_similarity(a, b):
    '''
    Equivalent to sklearn.metrics.pairwise.cosine_similarity
    '''
    a, b = torch.tensor(a), torch.tensor(b)
    similarity = torch.matmul(F.normalize(a), F.normalize(b).T)
    return similarity.numpy()


class CosineSimilarity(Similarity):
    def __call__(self, query, database):
        return {'default': cosine_similarity(query, database) }



import torch


class OptimizerAdam():
    '''
    Example config in YAML file:
    - Needs model params for initialization.

    optimizer:
        method: 'OptimizerAdam'
        lr: 1e-3
    '''

    @classmethod
    def from_config(cls, config, params):
        return torch.optim.Adam(params=params, **config)


class OptimizerSGD():
    '''
    Example config in YAML file:
    - Needs model params for initialization.

    optimizer:
        method: 'OptimizerSGD'
        lr: 1e-3
    '''

    @classmethod
    def from_config(cls, config, params):
        return torch.optim.SGD(params=params, **config)


class SchedulerCosine():
    '''
    Example config in YAML file:
    - Needs optimizer and number of epochs for initialization.

    scheduler:
        method: SchedulerCosine
    '''

    @classmethod
    def from_config(cls, config, optimizer, epochs):
        lr = optimizer.defaults.get("lr")
        lr_min = lr * 1e-3 if lr is not None else 1e-5
        return torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=epochs, eta_min=lr_min, **config)



import torch
import numpy as np
import pandas as pd
from collections import defaultdict



class NearestClassifier():
    def __call__(self, similarity, labels=None):
        '''
        Input is query x database similarity matrix and optionally mapping to database labels.
        '''
        similarity = torch.tensor(similarity, dtype=float)
        scores, idx = similarity.topk(k=1, dim=1)
        pred = pred.numpy().flatten()

        if labels is not None:
            pred = labels[pred]
        return pred




class KnnClassifier():

    def __init__(self, k: int = 1):
        self.k = k


    def __call__(self, similarity, labels=None):
        '''
        Input is query x database similarity matrix and optionally mapping to database labels.
        '''
        similarity = torch.tensor(similarity, dtype=float)
        scores, idx = similarity.topk(k=self.k, dim=1)
        pred = self.aggregate(idx)[:, self.k-1]

        if labels is not None:
            pred = labels[pred]
        return pred


    def aggregate(self, preds):
        '''
        Aggregates array of nearest neigbours to single prediction for each k.
        If there is tie at given k, prediction from k-1 is used.

        Input:
            - array of with shape [n_query, k] of nearest neighbours.

        Output: 
            - array of shape [n_query, k] of predicitons.
            - Column dimensions corresponds to a prediction for [k=1, k=2 ... k=k]
        '''

        results = defaultdict(list)
        for k in range(1, preds.shape[1] + 1):
            for row in preds[:, :k]:
                vals, counts = np.unique(row, return_counts=True)
                best = vals[np.argmax(counts)]

                counts_sorted = sorted(counts)
                if (len(counts_sorted)) > 1 and (counts_sorted[0] == counts_sorted[1]):
                    best = None
                results[k].append(best)

        results = pd.DataFrame(results).T.fillna(method='ffill').T
        return results.values






import pycocotools.mask as mask_coco
import cv2
import os
import numpy as np
import pandas as pd
from PIL import Image
import json
from wildlife_tools.tools import realize


class WildlifeDataset():
    '''
    PyTorch-style dataset for a wildlife image classification task.

    Args:
        metadata: A pandas dataframe containing image metadata.
        root: Root directory for images.
        transform: A function/transform that takes in an image and returns a transformed version.
        img_load: Method to load images. 
            Options: 'full', 'full_mask', 'full_hide', 'bbox', 'bbox_mask', 'bbox_hide', and 'crop_black'.
        col_path: Column name in the dataframe containing image file paths.
        col_identity: Column name in the dataframe containing class labels.

    Attributes:
        label: An array of integer labels for all images in the dataset.
        label_map: A mapping between integer label and string label.
            
    Methods:
        get_image(path): Load an image from file and convert it to a PIL Image object.
        
    Properties:
        num_classes: Return the number of unique classes in the dataset.
        __len__: Return the number of samples in the dataset.
        __getitem__(idx): Return (sample, label) tuple. Get a sample from the dataset at a given index.
    '''

    def __init__(
        self,
        metadata,
        root,
        split=None,
        transform=None,
        img_load='full',
        col_path='path',
        col_label='identity',
        load_label=True,
    ):
        self.split = split
        if self.split:
            metadata = self.split(metadata)

        self.metadata = metadata.reset_index(drop=True)
        self.root = root
        self.transform = transform
        self.img_load = img_load
        self.col_path = col_path
        self.col_label = col_label
        self.load_label = load_label
        self.labels, self.labels_map = pd.factorize(self.metadata[self.col_label].values)

    @property
    def labels_string(self):
        return self.metadata[self.col_label].astype(str).values

    @property
    def num_classes(self):
        return len(self.labels_map)


    def __len__(self):
        return len(self.metadata)


    def get_image(self, path):
        img = cv2.imread(path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(img)
        return img


    def __getitem__(self, idx):
        data = self.metadata.iloc[idx]
        img = self.get_image(os.path.join(self.root, data[self.col_path]))
        
        if self.img_load in ['full_mask', 'full_hide', 'bbox_mask', 'bbox_hide']:
            if not ('segmentation' in data):
                raise ValueError(f"{self.img_load} selected but no segmentation found.")
            if type(data['segmentation']) == str:
                segmentation = eval(data['segmentation'])
            else:
                segmentation = data['segmentation']

        if self.img_load in ['bbox', 'bbox_mask', 'bbox_mask']:
            if not ('bbox' in data):
                raise ValueError(f"{self.img_load} selected but no bbox found.")
            if type(data['bbox']) == str:
                bbox = json.loads(data['bbox'])
            else:
                bbox = data['bbox']

        # Load full image as it is.
        if self.img_load == 'full':
            img = img

        # Mask background using segmentation mask.
        elif self.img_load == 'full_mask':
            mask = mask_coco.decode(segmentation).astype('bool')
            img = Image.fromarray(img * mask[..., np.newaxis])
        
        # Hide object using segmentation mask
        elif self.img_load == 'full_hide':
            mask = mask_coco.decode(segmentation).astype('bool')
            img = Image.fromarray(img * ~mask[..., np.newaxis])

        # Crop to bounding box
        elif self.img_load == 'bbox':
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Mask background using segmentation mask and crop to bounding box.
        elif self.img_load == 'bbox_mask':
            mask = mask_coco.decode(segmentation).astype('bool')
            img = Image.fromarray(img * mask[..., np.newaxis])
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Hide object using segmentation mask and crop to bounding box.
        elif self.img_load == 'bbox_hide':
            mask = mask_coco.decode(segmentation).astype('bool')
            img = Image.fromarray(img * ~mask[..., np.newaxis])
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Crop black background around images
        elif self.img_load == 'crop_black':
            y_nonzero, x_nonzero, _ = np.nonzero(img)
            img = img.crop((np.min(x_nonzero), np.min(y_nonzero), np.max(x_nonzero), np.max(y_nonzero)))

        else:
            raise ValueError(f'Invalid img_load argument: {self.img_load}')

        if self.transform:
            img = self.transform(img)


        if self.load_label:
            return img, self.labels[idx]
        else:
            return img


    @classmethod
    def from_config(cls, config):
        config['split'] = realize(config.get('split'))
        config['transform'] = realize(config.get('transform'))
        config['metadata'] = pd.read_csv(config['metadata'], index_col=False)
        return cls(**config)

    


class FeatureDataset():
    def __init__(
        self,
        features,
        metadata,
        col_label='identity',
        load_label=True,
    ):

        if len(features) != len(metadata):
            raise ValueError('Features and metadata (lables) have different length ! ') 

        self.load_label = load_label
        self.features = features
        self.metadata = metadata.reset_index(drop=True)
        self.col_label=col_label
        self.labels, self.labels_map = pd.factorize(self.metadata[self.col_label].values)


    def save(self, path):
        data = {
            'features': self.features,
            'metadata': self.metadata,
        }
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as file:
            pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)


    @classmethod
    def load(cls, path, **config):
        with open(path, 'rb') as file:
            data = pickle.load(file)
        return cls(**data, **config)


    def __getitem__(self, idx):
        if self.load_label:
            return self.features[idx], self.labels[idx]
        else:
            return self.features[idx]


    def __len__(self):
        return len(self.metadata)


    @property
    def num_classes(self):
        return len(self.labels_map)


    @classmethod
    def from_config(cls, config):
        path = config.pop('path')
        return cls.load(path, **config)



import cv2
import numpy as np
from tqdm import tqdm
from wildlife_tools.features.base import FeatureExtractor


class SIFTFeatures(FeatureExtractor):
    def __init__(
        self,
        max_keypoints: int | None = None,
        descriptor_dim: int = 128,
    ):
        self.descriptor_dim = descriptor_dim
        self.max_keypoints = max_keypoints

    def __call__(self, dataset):
        if self.max_keypoints:
            sift = cv2.SIFT_create(nfeatures=self.max_keypoints)
        else:
            sift = cv2.SIFT_create()

        descriptors = []
        for img, y in tqdm(dataset, mininterval=1, ncols=100):
            keypoint, d = sift.detectAndCompute(np.array(img), None)
            if len(keypoint) <= 1:
                descriptors.append(None)
            else:
                descriptors.append(d)
        return descriptors



'''
from data.split import SplitChunk, SplitSplitter, SplitMetadata, SplitPipeline
from data.transform import TransformTimm, TransformTorchvision
from data.dataset import WildlifeDataset

split_store = {
    'SplitChunk': SplitChunk,
    'SplitSplitter': SplitSplitter,
    'SplitMetadata': SplitMetadata,
    'SplitPipeline': SplitPipeline,
}

transform_store = {
    'TransformTimm': TransformTimm,
    'TransformTorchvision': TransformTorchvision,
}


dataset_store = {
    'WildlifeDataset': WildlifeDataset,
    #'FeatureDataset': FeatureDataset,
}


'''


import torch
import os
import numpy as np
from tqdm import tqdm 
from wildlife_tools.features.base import FeatureExtractor


class InMemoryFeatures(FeatureExtractor):
    ''' Loads dataset in memory for faster access.'''

    def __call__(self, dataset):
        features = []
        for x, y in tqdm(dataset, mininterval=1, ncols=100):
            features.append(x)
        return features



import torch.nn as nn
from pytorch_metric_learning import losses, distances, miners


class ArcFaceLoss(nn.Module):
    ''' 
    Wraps Pytorch Metric Learning ArcFaceLoss

    Default margin (in radians): 0.5
    Default scale: 64
    '''
    
    def __init__(self, num_classes, embedding_size, margin=0.5, scale=64):
        super().__init__()
        self.loss = losses.ArcFaceLoss(
            num_classes=num_classes,
            embedding_size=embedding_size,
            margin=57.3*margin,
            scale=scale,
        )

    def forward(self, embeddings, y):
        return self.loss(embeddings, y)


class TripletLoss(nn.Module):
    ''' 
    Wraps Pytorch Metric Learning TripletMarginLoss

    Mining is one of: 'all', 'hard', 'semihard'
    Distance is one of: 'cosine', 'l2', 'l2_squared'
    '''
    def __init__(self, margin=0.2, mining='seminard', distance='l2_squared'):
        super().__init__()
        if distance == 'cosine':
            distance = distances.CosineSimilarity()
        elif distance == 'l2':
            distance = distances.LpDistance(normalize_embeddings=True, p=2, power=1)
        elif distance == 'l2_squared':
            distance = distances.LpDistance(normalize_embeddings=True, p=2, power=2)
        else:
            raise ValueError(f'Invalid distance: {distance}')


        self.loss = losses.TripletMarginLoss(distance=distance, margin=margin)
        self.miner = miners.TripletMarginMiner(distance=distance, type_of_triplets=mining, margin=margin)

    def forward(self, embeddings, y):
        indices_tuple = self.miner(embeddings, y)
        return self.loss(embeddings, y, indices_tuple)


class SoftmaxLoss(nn.Module):
    '''
    CE with simple single dense layer classification head.
    '''
    def __init__(self, num_classes, embedding_size):
        super().__init__()
        self.criterion = nn.CrossEntropyLoss()
        self.linear = nn.Linear(embedding_size, num_classes)

    def forward(self, x, y):
        return self.criterion(self.linear(x), y)



import torch
import torch.backends.cudnn
import torch.nn as nn
import random, os
import numpy as np
from tqdm import tqdm 
from itertools import chain
from copy import deepcopy
from wildlife_tools.tools import realize


def set_seed(seed=0, device='cuda'):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if device=='cuda':
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def set_random_states(states, device='cuda'):
    os.environ['PYTHONHASHSEED'] = states['os_rng_state']
    random.setstate(states["random_rng_state"])
    np.random.set_state(states["numpy_rng_state"])
    torch.set_rng_state(states["torch_rng_state"])
    if device=='cuda':
        torch.cuda.set_rng_state(states["torch_cuda_rng_state"])
        torch.cuda.set_rng_state_all(states["torch_cuda_rng_state_all"])
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_random_states(device='cuda'):
    states = {}
    states['os_rng_state'] = os.environ.get('PYTHONHASHSEED')
    states['random_rng_state'] = random.getstate()
    states['numpy_rng_state'] = np.random.get_state()
    states['torch_rng_state'] = torch.get_rng_state()
    if device=='cuda':
        states['torch_cuda_rng_state'] = torch.cuda.get_rng_state()
        states['torch_cuda_rng_state_all'] = torch.cuda.get_rng_state_all()
    return states


class BasicTrainer():
    def __init__(
        self,
        dataset,
        model,
        objective,
        optimizer,
        epochs,
        scheduler=None,
        device='cuda',
        batch_size=128,
        num_workers=1,
        accumulation_steps=1,
        epoch_callback=None,
    ):
        self.dataset = dataset
        self.model = model.to(device)
        self.objective = objective.to(device)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.epochs = epochs
        self.epoch = 0
        self.device = device
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.accumulation_steps = accumulation_steps
        self.epoch_callback = epoch_callback


    def train(self):
        loader = torch.utils.data.DataLoader(
            self.dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True
        )

        for e in range(self.epochs):
            epoch_data = self.train_epoch(loader)
            self.epoch += 1

            if self.epoch_callback:
                self.epoch_callback(trainer=self, epoch_data=epoch_data)


    def train_epoch(self, loader):
        model = self.model.train()
        for i, batch in tqdm(enumerate(loader), desc=f'Epoch {self.epoch}: ', mininterval=1):
            x, y = batch
            x, y = x.to(self.device), y.to(self.device)

            out = model(x)
            loss = self.objective(out, y)
            loss.backward()
            if (i-1) % self.accumulation_steps == 0:
                self.optimizer.step()
                self.optimizer.zero_grad()

            losses.append(loss.detach())

        if self.scheduler:
            self.scheduler.step()

        return {'train_loss_epoch_avg': losses.mean()}

    def save(self, folder, file_name='checkpoint.pth', **kwargs):
        if not os.path.exists(folder):
           os.makedirs(folder)
        if self.scheduler:
            scheduler_state = self.scheduler.state_dict()
        else:
            scheduler_state = None

        checkpoint = {
            'model': self.model.state_dict(),
            'objective': self.objective.state_dict(),
            'optimizer': self.optimizer.state_dict(),
            'epoch': self.epoch,
            'scheduler': scheduler_state,
            'rng_states': get_random_states(device=self.device),
        }
        torch.save(checkpoint, os.path.join(folder, file_name))


    def load(self, path, device='cpu'):
        checkpoint = torch.load(path, map_location=torch.device(device))

        if 'rng_states' in checkpoint:
            set_random_states(checkpoint['rng_states'], device=self.device)
        if 'model' in checkpoint:
            self.model.load_state_dict(checkpoint['model'])
        if 'objective' in checkpoint:
            self.objective.load_state_dict(checkpoint['objective'])
        if 'optimizer' in checkpoint:
            self.optimizer.load_state_dict(checkpoint['optimizer'])
        if 'epoch' in checkpoint:
            self.epoch = checkpoint['epoch']
        if 'scheduler' in checkpoint:
            self.scheduler.load_state_dict(checkpoint['scheduler'])



class ClassifierTrainer():
    @classmethod
    def from_config(cls, config):
        """
        Parses config dict setup BasicTrainer for training classifier.

        Config keys:
            dataset (dict):
                Config dictionary of the training dataset.
            backbone (dict):
                Config dictionary of the backbone.
            objective (dict):
                Config dictionary of the objective.
            scheduler (dict | None, default: None):
                Config dictionary of the scheduler (no scheduler is used by default).
            epochs (int):
                Number of training epochs.
            device (str, default: 'cuda'):
                Device to be used.
            batch_size (int, default: 128):
                Training batch size.
            num_workers (int, default: 1):
                Number of data loading workers in torch DataLoader.
            accumulation_steps (int, default: 1):
                Number of gradient accumulation steps.

        Returns:
            Ready to use BasicTrainer

        """

        config = deepcopy(config)

        dataset = realize(
            config=config.pop('dataset'),
        )
        model = realize(
            config=config.pop('backbone'),
            output_size=dataset.num_classes,
        )
        objective = realize(
            config=config.pop('objective'),
        )
        optimizer = realize(
            config=config.pop('optimizer'),
            params=model.parameters(),
        )
        scheduler = realize(
            config=config.pop('scheduler', None),
            epochs=config.get('epochs'),
        )
        epoch_callback=realize(
            config = config.pop('epoch_callback', None),
        )

        return BasicTrainer(
            model=model,
            dataset=dataset,
            objective=objective,
            optimizer=optimizer,
            scheduler=scheduler,
            epoch_callback=epoch_callback,
            **config
        )


class EmbeddingTrainer():
    @classmethod
    def from_config(cls, config):
        """
        Parses config dict setup BasicTrainer for training embedder.
    
        Config keys:
            dataset (dict):
                Config dictionary of the training dataset.
            backbone (dict):
                Config dictionary of the backbone.
            objective (dict):
                Config dictionary of the objective.
            scheduler (dict | None, default: None):
                Config dictionary of the scheduler (no scheduler is used by default).
            epochs (int):
                Number of training epochs.
            embedding_size (int | None, default: None):
                Adds a linear layer after the backbone with the target embedding size.
                By default, embedding size is inferred from the backbone (e.g., num_classes=0 in TIMM).
            device (str, default: 'cuda'):
                Device to be used.
            batch_size (int, default: 128):
                Training batch size.
            num_workers (int, default: 1):
                Number of data loading workers in torch DataLoader.
            accumulation_steps (int, default: 1):
                Number of gradient accumulation steps.

        Returns:
            Ready to use BasicTrainer

        """

        config = deepcopy(config)
        embedding_size = config.pop('embedding_size', None)
        
        dataset = realize(
            config=config.pop('dataset')
        )
        backbone = realize(
            config=config.pop('backbone'),
            output_size=embedding_size
        )

        if embedding_size is None: # Infers embedding size
            with torch.no_grad():
                x = dataset[0][0].unsqueeze(0)
                embedding_size = backbone(x).shape[1]

        objective = realize(
            config=config.pop('objective'),
            embedding_size=embedding_size,
            num_classes=dataset.num_classes,
        )
        optimizer = realize(
            config=config.pop('optimizer'),
            params=chain(backbone.parameters(), objective.parameters()),
        )
        scheduler = realize(
            optimizer=optimizer,
            config=config.pop('scheduler', None),
            epochs=config.get('epochs'),
        )
        epoch_callback=realize(
            config = config.pop('epoch_callback', None),
        )
        return BasicTrainer(
            model=backbone,
            dataset=dataset,
            objective=objective,
            optimizer=optimizer,
            scheduler=scheduler,
            epoch_callback=epoch_callback,
            **config
        )



import importlib
from copy import deepcopy
from jinja2 import Template, Environment, meta
import yaml
import torch
import torch.backends.cudnn
import torch.nn as nn
import random, os
import numpy as np

class Store():
    ''' Dictionary with components. Components are imported on demand to prevent circular imports.'''

    modules = {
        # Pipelines
        'SplitChunk': ('data.split', 'SplitChunk'),
        
        # Data
        'SplitChunk': ('wildlife_tools.data.split', 'SplitChunk'),
        'SplitWildlife': ('wildlife_tools.data.split', 'SplitWildlife'),
        'SplitMetadata': ('wildlife_tools.data.split', 'SplitMetadata'),
        'SplitPipeline': ('wildlife_tools.data.split', 'SplitPipeline'),
    
        'TransformTimm': ('wildlife_tools.data.transform', 'TransformTimm'),
        'TransformTorchvision': ('wildlife_tools.data.transform', 'TransformTorchvision'),

        'WildlifeDataset': ('wildlife_tools.data.dataset', 'WildlifeDataset'),
        'FeatureDataset': ('wildlife_tools.data.dataset', 'FeatureDataset'),


        # Features & Similarity
        'SimilarityPipeline': ('wildlife_tools.pipelines.similarity', 'SimilarityPipeline'),

        
        

        # Train
        'TrainingPipeline': ('wildlife_tools.pipelines.train', 'TrainingPipeline'),
        'EpochCallbacks': ('wildlife_tools.train.callbacks', 'EpochCallbacks'),
        'EpochLog': ('wildlife_tools.train.callbacks', 'EpochLog'),
        'EpochCheckpoint': ('wildlife_tools.train.callbacks', 'EpochCheckpoint'),


        'EmbeddingTrainer': ('wildlife_tools.train.trainer', 'EmbeddingTrainer'),
        'ClassifierTrainer': ('wildlife_tools.train.trainer', 'ClassifierTrainer'),

        'OptimizerAdam': ('wildlife_tools.train.optim', 'OptimizerAdam'),
        'OptimizerSGD': ('wildlife_tools.train.optim', 'OptimizerSGD'),
        'SchedulerCosine': ('wildlife_tools.train.optim', 'SchedulerCosine'),


        'TimmBackbone': ('wildlife_tools.train.backbone', 'TimmBackbone'),

        'ArcFaceLoss': ('wildlife_tools.train.objective', 'ArcFaceLoss'),
        'TripletLoss': ('wildlife_tools.train.objective', 'TripletLoss'),
        'SoftmaxLoss': ('wildlife_tools.train.objective', 'SoftmaxLoss'),
    }


    def __getitem__(self, key):
        module_path, class_name = self.modules[key]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)

    def __contains__(self, key):
        return key in self.modules


def realize(config, store=None, **kwargs):
    ''' realize object from config given its name and method store '''

    if config is None:
        return None

    if store is None:
        store = Store()

    config_local = deepcopy(config)
    name = config_local.pop('method', None)

    if name is None:
        raise ValueError(f"No 'method' attribute in config")
    if name not in store:
        raise ValueError(f"No method {name} in method store.")

    obj_class = store[name]
    if hasattr(obj_class, 'from_config'):
        obj = obj_class.from_config(config=config_local, **kwargs)
    else:
        obj = obj_class(**config_local, **kwargs)

    obj.source_config = deepcopy(config)
    return obj


def parse_yaml(yaml_string):
    '''
    Impute variables with in "{{ }}" with values from top level of the yaml dictionary.
    Example:

    yaml_string = """
    a:
        name: test
    b:
        uses: "{{a}}"
    """
    parse_yaml(yaml_string)
    >>> {'a': {'name': 'test'}, 'b': {'uses': {'name': 'test'}}}
    '''

    env = Environment(
        variable_start_string = "'{{",
        variable_end_string = "}}'",
    )

    data = yaml.safe_load(yaml_string)
    data_str = str(data)
    variables = meta.find_undeclared_variables(env.parse(data_str))

    while len(variables) != 0:
        data_str = env.from_string(data_str).render(data)
        data = yaml.safe_load(data_str)
        variables_new = meta.find_undeclared_variables(env.parse(data_str))

        # Check if number of resolved variables changed
        if len(variables) == len(variables_new):
            raise ValueError(f'Unable to impute variables in all places: {variables}')

        variables = variables_new
    return data


def set_seed(seed=0, device='cuda'):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if device == 'cuda':
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


import pandas as pd
from pandas.util.testing import assert_frame_equal

def test_sp_extractor(dataset):
    dataset.transform = T.Compose([
        T.Resize(size=256),
        T.CenterCrop(size=(224, 224)),
        T.Grayscale(),
        T.ToTensor(),
    ])
    extractor = SuperPointFeatures()
    features = extractor.get_features(dataset)

    assert_frame_equal(features.metadata, dataset.metadata)
    assert len(features.features) == len(features.metadata)
    assert [i.shape for i in features.features] == [(76, 256), (151, 256)]

test_sp_extractor(dataset)

def test_sift_extractor(dataset):
    dataset.transform = T.Compose([
        T.Resize(size=256),
        T.CenterCrop(size=(224, 224)),
        T.Grayscale(),
        ])
    extractor = SIFTFeatures()
    features = extractor.get_features(dataset)

    assert_frame_equal(features.metadata, dataset.metadata)
    assert len(features.features) == len(features.metadata)
    assert [i.shape for i in features.features] == [(209, 128), (427, 128)]

test_sift_extractor(dataset)



import torch
import os
import numpy as np
from tqdm import tqdm 
from wildlife_tools.features.base import FeatureExtractor
from wildlife_tools.tools import realize


class DeepFeatures(FeatureExtractor):
    def __init__(
        self,
        model,
        batch_size=128,
        num_workers=1,
        device='cpu',
    ):
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.device = device
        self.model = model


    def __call__(self, dataset):
        self.model = self.model.to(self.device)
        self.model = self.model.eval()

        loader = torch.utils.data.DataLoader(
            dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            shuffle=False,
        )
        outputs = []
        for image, label in tqdm(loader, mininterval=1, ncols=100):
            with torch.no_grad():
                output = self.model(image.to(self.device))
                outputs.append(output.cpu())
        return torch.cat(outputs).numpy()


    @classmethod
    def from_config(cls, config):
        model = realize(config.pop('model'))
        return cls(model=model, **config)